"""Functions for writing, given a samyro model."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy
import prettytensor as pt
import six

from samyro import integerize


def write(input_placeholder, output_activations,
          seed, max_length=1024, temperature=1.0):
    """Samples from the model

    Sampling is done by first running either the seed or an arbitrary character
    through the model and then drawing the next character from the probability
    distribution definted by `softmax`.

    Args:
      input_placeholder: tensorflow input placeholder, provided by model
      output_activations: tensorflow output activations tensor, provided by
        model
      seed: A non-empty string of characters to prime the network.
      max_length: The maximum length to draw in case EOS is not reached.
      temperature: A positive value used to renormalize the inputs.  A higher
        value selects less likely choices. 1.0 leaves the distribution alone.
    Returns:
      A string that was sampled from the model.
    """
    assert temperature > 0, 'Temperature must be greater than 0.'
    assert isinstance(seed, six.string_types) and len(seed)
    # The recurrent runner takes care of tracking the model's state at
    # each step and provides a reset call to zero it out for each query.
    recurrent_runner = pt.train.RecurrentRunner()

    # We need to reset the hidden state for each query.
    recurrent_runner.reset()
    # Initialize the system
    result = ''
    for c in seed[:-1]:
        recurrent_runner.run([output_activations],
                             {input_placeholder: integerize.convert_to_int(c)})
        result += c

    # Start sampling!
    ci = integerize.convert_to_int(seed[-1])
    while len(result) < max_length and ci != integerize.EOS:
        result += chr(ci)
        # The softmax is probability normalized and would have been
        # appropriate here if we weren't applying the temperature
        # (temperature could also be done in TensorFlow).
        logit_result = recurrent_runner.run([output_activations],
                                            {input_placeholder: ci})[0][0]
        logit_result /= temperature

        # Apply the softmax in numpy to convert from logits to
        # probabilities.  Subtract off the max for numerical stability
        # -- logits are invariant to additive scaling and this
        # eliminates overflows.
        logit_result -= logit_result.max()

        distribution = numpy.exp(logit_result)
        distribution /= distribution.sum()

        # Numpy multinomial needs the value to be strictly < 1
        distribution -= .00000001
        ci = numpy.argmax(numpy.random.multinomial(1, distribution))
    result += integerize.convert_to_symbol(ci)  # Add the last letter.
    return result.replace(
        integerize.BOS_CHAR, '').replace(integerize.EOS_CHAR, '')



"""Tzara learner parameters."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from samyro import read, model


class Trainer(object):
    """A training regime."""
    def __init__(self, model_instance, sampler, batch_size, reuse):

        assert isinstance(model_instance, model.Model)
        assert isinstance(sampler, read.Sampler)

        input_placeholder = sampler.placeholder()
        output_placeholder = sampler.placeholder()
        labels = sampler.as_labels(output_placeholder)

        self.feed_vars = (input_placeholder, output_placeholder)
        self.train_op, self.train_loss = model_instance.train_op_loss(
            input_placeholder,
            labels, reuse=reuse)
        self.eval_accuracy = model_instance.eval_accuracy(
            input_placeholder,
            labels, reuse=True)

    def train_pass(self, runner, batch_iterator, num_batches=300,
                   print_every=25):
        runner.train_model(train_op=self.train_op,
                           cost_to_log=self.train_loss,
                           num_steps=num_batches,
                           feed_vars=self.feed_vars,
                           feed_data=batch_iterator,
                           print_every=print_every)

    def eval_pass(self, runner, batch_iterator, num_batches=300):
        return runner.evaluate_model(accuracy=self.eval_accuracy,
                                     num_steps=num_batches,
                                     feed_vars=self.feed_vars,
                                     feed_data=batch_iterator)



from __future__ import print_function

import argparse
# import tensorflow as tf

import itertools

from samyro.cli import positive_int
import samyro.cli.shared
import samyro.integerize


def set_sampler_args(parser):
    samyro.cli.shared.set_sampler_shared_args(parser, default_batch_size=1)

    parser.add_argument('--shuffle', help="whether to shuffle samples",
                        default=True, type=bool)

    parser.add_argument('--batches', help="how many samples to draw",
                        default=5,
                        type=positive_int)

    parser.add_argument('--format', help="how to format examples",
                        choices=['text', 'numpy'],
                        default='text')

    # Positional argument for text file to read.
    parser.add_argument(
        'file', help="text file", type=argparse.FileType('r'),
        nargs="?",
        default='/opt/data/texts/shakespeare_input.txt')


def execute(args):
    """Actual execution of the sampler."""
    sampler = samyro.cli.shared.get_sampler(args)

    print("args: ", args)

    if args.format == 'text':
        stream = sampler.batches(shuffle=args.shuffle, to_numpy=False)
    elif args.format == 'numpy':
        stream = sampler.batches(shuffle=args.shuffle, to_numpy=True)
    else:
        assert False, "whoa nelly"

    for b_i, b in enumerate(itertools.islice(stream, args.batches)):
        print("# batch %s" % b_i)
        for i in b:
            print(i)


def main():
    """Entry point for main samyro-write script."""

    # shared_parent = samyro.cli.shared.set_shared_args()

    parser = argparse.ArgumentParser(prog='samyro-sample',
                                     fromfile_prefix_chars='@',)

    set_sampler_args(parser)

    args = parser.parse_args()

    execute(args)



from __future__ import print_function
import argparse

import tensorflow as tf

from samyro.cli import positive_int, positive_float

from six.moves import xrange

import samyro.cli.shared
import samyro.read
import samyro.learn


def execute(args):
    """The actual execution of learning (training)."""
    print(args)
    model = samyro.cli.shared.get_model(args)

    sampler = samyro.cli.shared.get_sampler(args)

    trainer = samyro.learn.Trainer(model_instance=model,
                                   sampler=sampler,
                                   batch_size=args.batch_size, reuse=None)

    with tf.Session():   # do we need to call out this sess variable?
        runner = samyro.cli.shared.get_runner(args)
        for epoch in xrange(args.epochs):
            trainer.train_pass(
                runner,
                print_every=args.print_every,
                batch_iterator=sampler.batches(shuffle=args.shuffle_train),
                num_batches=args.batches_per_epoch)

            classification_accuracy = trainer.eval_pass(
                runner,
                batch_iterator=sampler.batches(shuffle=args.shuffle_eval),
                num_batches=args.eval_per_epoch)
            print('Next character accuracy after epoch %d: %g%%'
                  % (epoch + 1, classification_accuracy * 100))

            if args.sampler == 'patches':
                seed = samyro.integerize.random_seed_char()
            else:
                seed = samyro.integerize.BOS_CHAR
            print(samyro.write.write(*(model.inference_io(reuse=True)),
                                     max_length=args.sample_length,
                                     seed=seed,
                                     temperature=args.sample_temperature))


def set_learner_args(learner_parser):
    """samyro learn uses these arguments and defines the learn function."""

    samyro.cli.shared.set_sampler_shared_args(learner_parser)

    # Learner-specific arguments
    regime_group = learner_parser.add_argument_group(
        'training regime configuration')
    regime_group.add_argument('--epochs',
                              type=positive_int,
                              default=10,
                              help="how many training epochs to run")
    regime_group.add_argument('--batches_per_epoch',
                              type=positive_int,
                              default=300,
                              help="how many batches per epoch")
    regime_group.add_argument('--shuffle_train', default=True,
                              type=bool,
                              help="if each epoch should be a new random set")
    regime_group.add_argument(
        '--print_every', default=25,
        type=int,
        help="how many batches between training loss updates")

    eval_group = learner_parser.add_argument_group(
        'eval regime configuration: 1x per epoch')
    eval_group.add_argument('--eval_per_epoch',
                            type=positive_int,
                            default=300,
                            help="how many eval batches per epoch")
    eval_group.add_argument('--shuffle_eval', default=False,
                            type=bool,
                            help="if each eval should be a new random draw")

    sample_group = learner_parser.add_argument_group(
        'draw a sample from the model every epoch')
    sample_group.add_argument('--sample_length',
                              type=positive_int,
                              default=128,
                              help="max length of eval sample")
    sample_group.add_argument('--sample_temperature',
                              type=positive_float,
                              default=0.5,
                              help="temperature for eval samples")

    # Positional argument for text file to read.
    learner_parser.add_argument(
        'file', help="text file", type=argparse.FileType('r'),
        nargs="?",
        default='/opt/data/texts/shakespeare_input.txt')

    learner_parser.set_defaults(execute=execute)


def main():
    """Entry point for main samyro-learn script."""

    shared_parent = samyro.cli.shared.set_shared_args()

    parser = argparse.ArgumentParser(prog='samyro-learn',
                                     fromfile_prefix_chars='@',
                                     parents=[shared_parent])

    set_learner_args(parser)

    args = parser.parse_args()

    execute(args)



"""Test reading batches of character streams."""
from six.moves import cStringIO as StringIO
from samyro import integerize, read

import numpy

import re


def repair(string):
    string = re.sub(r'<BOS>', integerize.BOS_CHAR, string)
    string = re.sub(r'<EOS>', integerize.EOS_CHAR, string)
    return string


def test_patch_batches():
    rr = read.CharacterPatchSampler(filehandle=StringIO(u'foobar\n'),
                                    sample_length=3, batch_size=2)
    batches = rr.batches(shuffle=False)

    first = next(batches)
    assert isinstance(first, read.Sample)
    assert isinstance(first.input, numpy.ndarray)

    batches = rr.batches(shuffle=False, to_numpy=False)

    first = list(next(batches))
    assert len(first) == 2
    assert first[0] == read.Sample(input=repair("<BOS>fo"), output="foo")


def test_patch_strings():
    rr = read.CharacterPatchSampler(filehandle=StringIO(u'foobar\n'),
                                    sample_length=3, batch_size=2)
    strings = rr.string_samples()

    first = next(strings)
    assert isinstance(first, read.Sample)
    assert read.Sample(input=repair("<BOS>fo"), output="foo") == first

    second = next(strings)
    assert isinstance(second, read.Sample)
    assert read.Sample(input="oba", output="bar") == second

    third = next(strings)
    assert isinstance(third, read.Sample)
    assert read.Sample(input=repair("r\n<EOS>"),
                       output=repair("\n<EOS><EOS>")) == third

    # assert raises EndOfInput thereafter?


def test_lines_strings():
    rr = read.LineSampler(filehandle=StringIO(u'foobar\nbaz\nwikiwiki\n'),
                          sample_length=6, batch_size=2)
    strings = rr.string_samples()

    first = next(strings)
    assert isinstance(first, read.Sample)
    assert read.Sample(input=repair("<BOS>fooba"), output="foobar") == first

    second = next(strings)
    assert isinstance(second, read.Sample)
    assert read.Sample(input=repair("<BOS>baz\n<EOS>"),
                       output=repair("baz\n<EOS><EOS>")) == second

    third = next(strings)
    assert isinstance(third, read.Sample)
    assert read.Sample(input=repair("<BOS>wikiw"), output="wikiwi") == third

    # assert raises EndOfInput thereafter?


def test_paras_strings():
    rr = read.ParagraphSampler(
        filehandle=StringIO(u'foobar\nbaz\n\n\nwikiwiki\nyap\n\nwooo'),
        sample_length=15, batch_size=2)
    strings = rr.string_samples()

    first = next(strings)
    assert isinstance(first, read.Sample)
    assert read.Sample(
        input=repair("<BOS>foobar\nbaz\n<EOS><EOS><EOS>"),
        output=repair("foobar\nbaz\n<EOS><EOS><EOS><EOS>")) == first

    second = next(strings)
    assert isinstance(second, read.Sample)
    assert read.Sample(input=repair("<BOS>wikiwiki\nyap\n<EOS>"),
                       output=repair("wikiwiki\nyap\n<EOS><EOS>")) == second

    third = next(strings)
    assert isinstance(third, read.Sample)
    assert read.Sample(input=repair("<BOS>wooo" + "<EOS>" * 10),
                       output=repair("wooo" + "<EOS>" * 11)) == third

    # assert raises EndOfInput thereafter?



"""Tzara functions for loading up training batches from input text."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import codecs
import collections
import random

import numpy

import tensorflow as tf

from six.moves import zip_longest

# TODO(jkahn): remove dependency
from prettytensor.tutorial import data_utils

from samyro import integerize


class Sample(collections.namedtuple('Sample', ['input', 'output'])):
    """Contains an input/output pair (any format)."""
    @classmethod
    def to_numpy_batch(cls, batch, dtype=numpy.int32):
        """Takes an iterable of Sample objects; produces a numpy rectangle."""
        # assert that lengths are all the same?
        batch = [s.to_numpy_arrays(dtype=dtype) for s in batch]
        return cls(input=numpy.array([s.input for s in batch],
                                     dtype=dtype),
                   output=numpy.array([s.output for s in batch],
                                      dtype=dtype))

    def to_numpy_arrays(self, dtype=numpy.int32):
        """Returns numpy-ified version of string inputs for a single sample."""
        return Sample(input=numpy.array([integerize.convert_to_int(c)
                                         for c in self.input], dtype=dtype),
                      output=numpy.array([integerize.convert_to_int(c)
                                          for c in self.output], dtype=dtype))

    @classmethod
    def from_text(cls, text, length, right_pad=integerize.EOS_CHAR):
        # Pad out end of line with EOS_CHAR
        if len(text) < length:
            n_pad_chars = length - len(text)
            text += (right_pad * n_pad_chars)

        # Trim back overlong lines
        if len(text) > length:
            text = text[:length]

        # Apply offset, padding with BOS_CHAR
        return cls(input=integerize.BOS_CHAR + text[:-1],
                   output=text)

    def pretty(self):
        raise NotImplementedError("implement aligned output")


# TODO(jkahn): make this an ABC?
class Sampler(object):
    """An ABC for any sampler."""
    def __init__(self, sample_length, batch_size,
                 dummy_value=integerize.EOS_CHAR):
        self.sample_length = sample_length
        self.batch_size = batch_size
        self.dummy_value = dummy_value

    def __iter__(self):
        return self.stream()

    def string_samples(self):
        raise NotImplementedError("This is an abstract base class."
                                  " Implement 'string_samples' in subclasses.")

    def stream(self):
        for s in self.string_samples():
            yield s.to_numpy_arrays()

    @property
    def dummy_sequence(self):
        return Sample(input=self.dummy_value * self.sample_length,
                      output=self.dummy_value * self.sample_length)

    def batches(self, shuffle, to_numpy=True):
        seqs = self.string_samples()
        if shuffle:
            seqs = list(seqs)
            random.shuffle(seqs)
        batches = zip_longest(fillvalue=self.dummy_sequence,
                              *([iter(seqs)] * self.batch_size))
        if to_numpy:
            return (Sample.to_numpy_batch(b) for b in batches)
        else:
            return batches

    def placeholder(self):
        return integerize.placeholder(self.batch_size, self.sample_length)

    def as_labels(self, output_placeholder):
        tokens_per_batch = self.batch_size * self.sample_length
        _t = tf.concat(1,
                       [tf.constant(numpy.arange(tokens_per_batch).reshape(
                           (tokens_per_batch, 1)), dtype=tf.int32),
                        data_utils.reshape_data(output_placeholder)])
        return tf.sparse_to_dense(_t, [tokens_per_batch,
                                       integerize.CHARS],
                                  1.0, 0.0)


class FileSampler(Sampler):
    """File(handle) sampler base class."""
    def __init__(self, sample_length, batch_size,
                 encoding='utf-8',
                 filename=None, filehandle=None):
        assert filename is not None or filehandle is not None
        assert not all([filename, filehandle])
        if not filehandle:
            filehandle = codecs.open(filename, 'rb', encoding)
        self.filehandle = filehandle
        super(FileSampler, self).__init__(
            sample_length=sample_length, batch_size=batch_size,
            dummy_value=integerize.EOS_CHAR)

    @property
    def characters(self):
        """Returns iterable of characters."""
        self.filehandle.seek(0)
        return self.filehandle.read()

    @property
    def lines(self):
        """Returns iterable of lines.

        Default implementation uses file's linebreaks.
        """
        self.filehandle.seek(0)
        return self.filehandle

    # TODO(jkahn): make paragraph_sep into a regex
    def paragraphs(self, paragraph_sep=""):
        """Returns iterable of non-empty line groups.

        File is thereby separated by empty lines.
        """

        self.filehandle.seek(0)
        current = []
        for line in self.lines:
            if line.strip() == paragraph_sep:
                if current:
                    yield ''.join(current)
                    current = []
            else:
                current.append(line)
        if current:
            yield ''.join(current)


class LineSampler(FileSampler):
    """A file(handle) sampler that returns fixed-length lines.

    Each sample is trimmed or padded to the target line length.
    Each sample pads the leftmost input with one integerize.BOS_CHAR token
    and pads the right input with integerize.EOS_CHAR.

    Whitespace-only lines are ignored.
    """
    def string_samples(self):
        for line in self.lines:
            if len(line.strip()):
                yield Sample.from_text(line, length=self.sample_length)


class ParagraphSampler(FileSampler):
    """A file(handle) sampler that returns fixed-length paragraph samples.

    Same padding rules as LineSampler, but inputs are paragraphs instead.
    """
    def string_samples(self):
        for line in self.paragraphs():
            yield Sample.from_text(line, length=self.sample_length)


class CharacterPatchSampler(FileSampler):
    """A file(handle) sampler that returns patches of given sample_length.

    Input file uses integerize.BOS_CHAR as leftmost input token for
    first sample, but all other samples use an input and an output
    with an offset of 1.
    """
    def string_samples(self):
        char_iter = self.characters
        prev = integerize.BOS_CHAR
        for b in zip_longest(
                fillvalue=integerize.EOS_CHAR,
                *([iter(char_iter)] * self.sample_length)):
            yield Sample(input=prev + ''.join(b[:-1]),
                         output=''.join(b))
            prev = b[-1]
        if prev != integerize.EOS_CHAR:
            # last sample not rounded evenly
            yield Sample(
                input=prev + integerize.EOS_CHAR * (self.sample_length - 1),
                output=integerize.EOS_CHAR * self.sample_length)

# TODO(jeremy): generate training examples from each *line* or perhaps sliding
# pair thereof. Consider weighting the second line 1.0 but the first 0.0?
# What happens when we give it couplets?
